@@ -6,22 +6,24 @@ part of dart._http;
 
 class _HttpHeaders implements HttpHeaders {
   final Map<String, List<String>> _headers;
+  // The original header names keyed by the lowercase header names.
+  Map<String, String>? _originalHeaderNames;
   final String protocolVersion;
 
   bool _mutable = true; // Are the headers currently mutable?
-  List<String> _noFoldingHeaders;
+  List<String>? _noFoldingHeaders;
 
   int _contentLength = -1;
   bool _persistentConnection = true;
   bool _chunkedTransferEncoding = false;
-  String _host;
-  int _port;
+  String? _host;
+  int? _port;
 
   final int _defaultPortForScheme;
 
   _HttpHeaders(this.protocolVersion,
-      {int defaultPortForScheme: HttpClient.defaultHttpPort,
-      _HttpHeaders initialHeaders})
+      {int defaultPortForScheme = HttpClient.defaultHttpPort,
+      _HttpHeaders? initialHeaders})
       : _headers = new HashMap<String, List<String>>(),
         _defaultPortForScheme = defaultPortForScheme {
     if (initialHeaders != null) {
@@ -38,25 +40,32 @@ class _HttpHeaders implements HttpHeaders {
     }
   }
 
-  List<String> operator [](String name) => _headers[name.toLowerCase()];
+  List<String>? operator [](String name) => _headers[_validateField(name)];
 
-  String value(String name) {
-    name = name.toLowerCase();
-    List<String> values = _headers[name];
+  String? value(String name) {
+    name = _validateField(name);
+    List<String>? values = _headers[name];
     if (values == null) return null;
+    assert(values.isNotEmpty);
     if (values.length > 1) {
       throw new HttpException("More than one value for header $name");
     }
     return values[0];
   }
 
-  void add(String name, value) {
+  void add(String name, value, {bool preserveHeaderCase = false}) {
     _checkMutable();
-    _addAll(_validateField(name), value);
+    String lowercaseName = _validateField(name);
+
+    if (preserveHeaderCase && name != lowercaseName) {
+      (_originalHeaderNames ??= {})[lowercaseName] = name;
+    } else {
+      _originalHeaderNames?.remove(lowercaseName);
+    }
+    _addAll(lowercaseName, value);
   }
 
   void _addAll(String name, value) {
-    assert(name == _validateField(name));
     if (value is Iterable) {
       for (var v in value) {
         _add(name, _validateValue(v));
@@ -66,27 +75,34 @@ class _HttpHeaders implements HttpHeaders {
     }
   }
 
-  void set(String name, Object value) {
+  void set(String name, Object value, {bool preserveHeaderCase = false}) {
     _checkMutable();
-    name = _validateField(name);
-    _headers.remove(name);
-    if (name == HttpHeaders.transferEncodingHeader) {
+    String lowercaseName = _validateField(name);
+    _headers.remove(lowercaseName);
+    _originalHeaderNames?.remove(lowercaseName);
+    if (lowercaseName == HttpHeaders.contentLengthHeader) {
+      _contentLength = -1;
+    }
+    if (lowercaseName == HttpHeaders.transferEncodingHeader) {
       _chunkedTransferEncoding = false;
     }
-    _addAll(name, value);
+    if (preserveHeaderCase && name != lowercaseName) {
+      (_originalHeaderNames ??= {})[lowercaseName] = name;
+    }
+    _addAll(lowercaseName, value);
   }
 
   void remove(String name, Object value) {
     _checkMutable();
     name = _validateField(name);
     value = _validateValue(value);
-    List<String> values = _headers[name];
+    List<String>? values = _headers[name];
     if (values != null) {
-      int index = values.indexOf(value);
-      if (index != -1) {
-        values.removeRange(index, index + 1);
+      values.remove(_valueToString(value));
+      if (values.length == 0) {
+        _headers.remove(name);
+        _originalHeaderNames?.remove(name);
       }
-      if (values.length == 0) _headers.remove(name);
     }
     if (name == HttpHeaders.transferEncodingHeader && value == "chunked") {
       _chunkedTransferEncoding = false;
@@ -97,15 +113,19 @@ class _HttpHeaders implements HttpHeaders {
     _checkMutable();
     name = _validateField(name);
     _headers.remove(name);
+    _originalHeaderNames?.remove(name);
   }
 
-  void forEach(void f(String name, List<String> values)) {
-    _headers.forEach(f);
+  void forEach(void action(String name, List<String> values)) {
+    _headers.forEach((String name, List<String> values) {
+      String originalName = _originalHeaderName(name);
+      action(originalName, values);
+    });
   }
 
   void noFolding(String name) {
-    if (_noFoldingHeaders == null) _noFoldingHeaders = new List<String>();
-    _noFoldingHeaders.add(name);
+    name = _validateField(name);
+    (_noFoldingHeaders ??= <String>[]).add(name);
   }
 
   bool get persistentConnection => _persistentConnection;
@@ -113,20 +133,21 @@ class _HttpHeaders implements HttpHeaders {
   void set persistentConnection(bool persistentConnection) {
     _checkMutable();
     if (persistentConnection == _persistentConnection) return;
+    final originalName = _originalHeaderName(HttpHeaders.connectionHeader);
     if (persistentConnection) {
       if (protocolVersion == "1.1") {
         remove(HttpHeaders.connectionHeader, "close");
       } else {
-        if (_contentLength == -1) {
+        if (_contentLength < 0) {
           throw new HttpException(
               "Trying to set 'Connection: Keep-Alive' on HTTP 1.0 headers with "
               "no ContentLength");
         }
-        add(HttpHeaders.connectionHeader, "keep-alive");
+        add(originalName, "keep-alive", preserveHeaderCase: true);
       }
     } else {
       if (protocolVersion == "1.1") {
-        add(HttpHeaders.connectionHeader, "close");
+        add(originalName, "close", preserveHeaderCase: true);
       } else {
         remove(HttpHeaders.connectionHeader, "keep-alive");
       }
@@ -151,7 +172,7 @@ class _HttpHeaders implements HttpHeaders {
       if (chunkedTransferEncoding) chunkedTransferEncoding = false;
       _set(HttpHeaders.contentLengthHeader, contentLength.toString());
     } else {
-      removeAll(HttpHeaders.contentLengthHeader);
+      _headers.remove(HttpHeaders.contentLengthHeader);
       if (protocolVersion == "1.1") {
         chunkedTransferEncoding = true;
       }
@@ -168,8 +189,8 @@ class _HttpHeaders implements HttpHeaders {
     }
     if (chunkedTransferEncoding == _chunkedTransferEncoding) return;
     if (chunkedTransferEncoding) {
-      List<String> values = _headers[HttpHeaders.transferEncodingHeader];
-      if ((values == null || values.last != "chunked")) {
+      List<String>? values = _headers[HttpHeaders.transferEncodingHeader];
+      if (values == null || !values.contains("chunked")) {
         // Headers does not specify chunked encoding - add it if set.
         _addValue(HttpHeaders.transferEncodingHeader, "chunked");
       }
@@ -181,81 +202,96 @@ class _HttpHeaders implements HttpHeaders {
     _chunkedTransferEncoding = chunkedTransferEncoding;
   }
 
-  String get host => _host;
+  String? get host => _host;
 
-  void set host(String host) {
+  void set host(String? host) {
     _checkMutable();
     _host = host;
     _updateHostHeader();
   }
 
-  int get port => _port;
+  int? get port => _port;
 
-  void set port(int port) {
+  void set port(int? port) {
     _checkMutable();
     _port = port;
     _updateHostHeader();
   }
 
-  DateTime get ifModifiedSince {
-    List<String> values = _headers[HttpHeaders.ifModifiedSinceHeader];
+  DateTime? get ifModifiedSince {
+    List<String>? values = _headers[HttpHeaders.ifModifiedSinceHeader];
     if (values != null) {
+      assert(values.isNotEmpty);
       try {
         return HttpDate.parse(values[0]);
-      } on Exception catch (e) {
+      } on Exception {
         return null;
       }
     }
     return null;
   }
 
-  void set ifModifiedSince(DateTime ifModifiedSince) {
+  void set ifModifiedSince(DateTime? ifModifiedSince) {
     _checkMutable();
-    // Format "ifModifiedSince" header with date in Greenwich Mean Time (GMT).
-    String formatted = HttpDate.format(ifModifiedSince.toUtc());
-    _set(HttpHeaders.ifModifiedSinceHeader, formatted);
+    if (ifModifiedSince == null) {
+      _headers.remove(HttpHeaders.ifModifiedSinceHeader);
+    } else {
+      // Format "ifModifiedSince" header with date in Greenwich Mean Time (GMT).
+      String formatted = HttpDate.format(ifModifiedSince.toUtc());
+      _set(HttpHeaders.ifModifiedSinceHeader, formatted);
+    }
   }
 
-  DateTime get date {
-    List<String> values = _headers[HttpHeaders.dateHeader];
+  DateTime? get date {
+    List<String>? values = _headers[HttpHeaders.dateHeader];
     if (values != null) {
+      assert(values.isNotEmpty);
       try {
         return HttpDate.parse(values[0]);
-      } on Exception catch (e) {
+      } on Exception {
         return null;
       }
     }
     return null;
   }
 
-  void set date(DateTime date) {
+  void set date(DateTime? date) {
     _checkMutable();
-    // Format "DateTime" header with date in Greenwich Mean Time (GMT).
-    String formatted = HttpDate.format(date.toUtc());
-    _set("date", formatted);
+    if (date == null) {
+      _headers.remove(HttpHeaders.dateHeader);
+    } else {
+      // Format "DateTime" header with date in Greenwich Mean Time (GMT).
+      String formatted = HttpDate.format(date.toUtc());
+      _set(HttpHeaders.dateHeader, formatted);
+    }
   }
 
-  DateTime get expires {
-    List<String> values = _headers[HttpHeaders.expiresHeader];
+  DateTime? get expires {
+    List<String>? values = _headers[HttpHeaders.expiresHeader];
     if (values != null) {
+      assert(values.isNotEmpty);
       try {
         return HttpDate.parse(values[0]);
-      } on Exception catch (e) {
+      } on Exception {
         return null;
       }
     }
     return null;
   }
 
-  void set expires(DateTime expires) {
+  void set expires(DateTime? expires) {
     _checkMutable();
-    // Format "Expires" header with date in Greenwich Mean Time (GMT).
-    String formatted = HttpDate.format(expires.toUtc());
-    _set(HttpHeaders.expiresHeader, formatted);
+    if (expires == null) {
+      _headers.remove(HttpHeaders.expiresHeader);
+    } else {
+      // Format "Expires" header with date in Greenwich Mean Time (GMT).
+      String formatted = HttpDate.format(expires.toUtc());
+      _set(HttpHeaders.expiresHeader, formatted);
+    }
   }
 
-  ContentType get contentType {
-    var values = _headers["content-type"];
+  ContentType? get contentType {
+    var values = _headers[HttpHeaders.contentTypeHeader];
     if (values != null) {
       return ContentType.parse(values[0]);
     } else {
@@ -263,9 +299,13 @@ class _HttpHeaders implements HttpHeaders {
     }
   }
 
-  void set contentType(ContentType contentType) {
+  void set contentType(ContentType? contentType) {
     _checkMutable();
-    _set(HttpHeaders.contentTypeHeader, contentType.toString());
+    if (contentType == null) {
+      _headers.remove(HttpHeaders.contentTypeHeader);
+    } else {
+      _set(HttpHeaders.contentTypeHeader, contentType.toString());
+    }
   }
 
   void clear() {
@@ -396,7 +436,7 @@ class _HttpHeaders implements HttpHeaders {
         } else {
           try {
             _port = int.parse(value.substring(pos + 1));
-          } on FormatException catch (e) {
+          } on FormatException {
             _port = null;
           }
         }
@@ -422,42 +462,41 @@ class _HttpHeaders implements HttpHeaders {
   }
 
   void _addValue(String name, Object value) {
-    List<String> values = _headers[name];
-    if (values == null) {
-      values = new List<String>();
-      _headers[name] = values;
-    }
+    List<String> values = (_headers[name] ??= <String>[]);
+    values.add(_valueToString(value));
+  }
+
+  String _valueToString(Object value) {
     if (value is DateTime) {
-      values.add(HttpDate.format(value));
+      return HttpDate.format(value);
     } else if (value is String) {
-      values.add(value);
+      return value; // TODO(39784): no _validateValue?
     } else {
-      values.add(_validateValue(value.toString()));
+      return _validateValue(value.toString()) as String;
     }
   }
 
   void _set(String name, String value) {
     assert(name == _validateField(name));
-    List<String> values = new List<String>();
-    _headers[name] = values;
-    values.add(value);
+    _headers[name] = <String>[value];
   }
 
-  _checkMutable() {
+  void _checkMutable() {
     if (!_mutable) throw new HttpException("HTTP headers are not mutable");
   }
 
-  _updateHostHeader() {
-    bool defaultPort = _port == null || _port == _defaultPortForScheme;
-    _set("host", defaultPort ? host : "$host:$_port");
+  void _updateHostHeader() {
+    var host = _host;
+    if (host != null) {
+      bool defaultPort = _port == null || _port == _defaultPortForScheme;
+      _set("host", defaultPort ? host : "$host:$_port");
+    }
   }
 
-  _foldHeader(String name) {
-    if (name == HttpHeaders.setCookieHeader ||
-        (_noFoldingHeaders != null && _noFoldingHeaders.indexOf(name) != -1)) {
-      return false;
-    }
-    return true;
+  bool _foldHeader(String name) {
+    if (name == HttpHeaders.setCookieHeader) return false;
+    var noFoldingHeaders = _noFoldingHeaders;
+    return noFoldingHeaders == null || !noFoldingHeaders.contains(name);
   }
 
   void _finalize() {
@@ -465,10 +504,10 @@ class _HttpHeaders implements HttpHeaders {
   }
 
   void _build(BytesBuilder builder) {
-    for (String name in _headers.keys) {
-      List<String> values = _headers[name];
+    _headers.forEach((String name, List<String> values) {
+      String originalName = _originalHeaderName(name);
       bool fold = _foldHeader(name);
-      var nameData = name.codeUnits;
+      var nameData = originalName.codeUnits;
       builder.add(nameData);
       builder.addByte(_CharCode.COLON);
       builder.addByte(_CharCode.SP);
@@ -489,20 +528,21 @@ class _HttpHeaders implements HttpHeaders {
       }
       builder.addByte(_CharCode.CR);
       builder.addByte(_CharCode.LF);
-    }
+    });
   }
 
   String toString() {
     StringBuffer sb = new StringBuffer();
     _headers.forEach((String name, List<String> values) {
-      sb..write(name)..write(": ");
+      String originalName = _originalHeaderName(name);
+      sb..write(originalName)..write(": ");
       bool fold = _foldHeader(name);
       for (int i = 0; i < values.length; i++) {
         if (i > 0) {
           if (fold) {
             sb.write(", ");
           } else {
-            sb..write("\n")..write(name)..write(": ");
+            sb..write("\n")..write(originalName)..write(": ");
           }
         }
         sb.write(values[i]);
@@ -514,7 +554,7 @@ class _HttpHeaders implements HttpHeaders {
 
   List<Cookie> _parseCookies() {
     // Parse a Cookie header value according to the rules in RFC 6265.
-    var cookies = new List<Cookie>();
+    var cookies = <Cookie>[];
     void parseCookieString(String s) {
       int index = 0;
 
@@ -577,7 +617,7 @@ class _HttpHeaders implements HttpHeaders {
       }
     }
 
-    List<String> values = _headers[HttpHeaders.cookieHeader];
+    List<String>? values = _headers[HttpHeaders.cookieHeader];
     if (values != null) {
       values.forEach((headerValue) => parseCookieString(headerValue));
     }
@@ -588,39 +628,46 @@ class _HttpHeaders implements HttpHeaders {
     for (var i = 0; i < field.length; i++) {
       if (!_HttpParser._isTokenChar(field.codeUnitAt(i))) {
         throw new FormatException(
-            "Invalid HTTP header field name: ${json.encode(field)}");
+            "Invalid HTTP header field name: ${json.encode(field)}", field, i);
       }
     }
     return field.toLowerCase();
   }
 
-  static _validateValue(value) {
+  static Object _validateValue(Object value) {
     if (value is! String) return value;
-    for (var i = 0; i < value.length; i++) {
-      if (!_HttpParser._isValueChar(value.codeUnitAt(i))) {
+    for (var i = 0; i < (value as String).length; i++) {
+      if (!_HttpParser._isValueChar((value as String).codeUnitAt(i))) {
         throw new FormatException(
-            "Invalid HTTP header field value: ${json.encode(value)}");
+            "Invalid HTTP header field value: ${json.encode(value)}", value, i);
       }
     }
     return value;
   }
+
+  String _originalHeaderName(String name) {
+    // TODO: Update syntax to_originalHeaderNames?[name].
+    return _originalHeaderNames?.[name] ?? name;
+  }
 }
 
 class _HeaderValue implements HeaderValue {
   String _value;
-  Map<String, String> _parameters;
-  Map<String, String> _unmodifiableParameters;
+  Map<String, String?>? _parameters;
+  Map<String, String?>? _unmodifiableParameters;
 
-  _HeaderValue([this._value = "", Map<String, String> parameters]) {
-    if (parameters != null) {
-      _parameters = new HashMap<String, String>.from(parameters);
+  _HeaderValue([this._value = "", Map<String, String?> parameters = const {}]) {
+    // TODO(40614): Remove once non-nullability is sound.
+    Map<String, String?>? nullableParameters = parameters;
+    if (nullableParameters != null && nullableParameters.isNotEmpty) {
+      _parameters = new HashMap<String, String?>.from(nullableParameters);
     }
   }
 
   static _HeaderValue parse(String value,
-      {parameterSeparator: ";",
-      valueSeparator: null,
-      preserveBackslash: false}) {
+      {String parameterSeparator = ";",
+      String? valueSeparator,
+      bool preserveBackslash = false}) {
     // Parse the string.
     var result = new _HeaderValue();
     result._parse(value, parameterSeparator, valueSeparator, preserveBackslash);
@@ -629,32 +676,61 @@ class _HeaderValue implements HeaderValue {
 
   String get value => _value;
 
-  void _ensureParameters() {
-    if (_parameters == null) {
-      _parameters = new HashMap<String, String>();
-    }
-  }
+  Map<String, String?> _ensureParameters() =>
+      _parameters ??= <String, String?>{};
 
-  Map<String, String> get parameters {
-    _ensureParameters();
-    if (_unmodifiableParameters == null) {
-      _unmodifiableParameters = new UnmodifiableMapView(_parameters);
+  Map<String, String?> get parameters =>
+      _unmodifiableParameters ??= UnmodifiableMapView(_ensureParameters());
+
+  static bool _isToken(String token) {
+    if (token.isEmpty) {
+      return false;
     }
-    return _unmodifiableParameters;
+    final delimiters = "\"(),/:;<=>?@[\]{}";
+    for (int i = 0; i < token.length; i++) {
+      int codeUnit = token.codeUnitAt(i);
+      if (codeUnit <= 32 ||
+          codeUnit >= 127 ||
+          delimiters.indexOf(token[i]) >= 0) {
+        return false;
+      }
+    }
+    return true;
   }
 
   String toString() {
     StringBuffer sb = new StringBuffer();
     sb.write(_value);
+    var parameters = this._parameters;
     if (parameters != null && parameters.length > 0) {
-      _parameters.forEach((String name, String value) {
-        sb..write("; ")..write(name)..write("=")..write(value);
+      parameters.forEach((String name, String? value) {
+        sb..write("; ")..write(name);
+        if (value != null) {
+          sb.write("=");
+          if (_isToken(value)) {
+            sb.write(value);
+          } else {
+            sb.write('"');
+            int start = 0;
+            for (int i = 0; i < value.length; i++) {
+              // Can use codeUnitAt here instead.
+              int codeUnit = value.codeUnitAt(i);
+              if (codeUnit == 92 /* backslash */ ||
+                  codeUnit == 34 /* double quote */) {
+                sb.write(value.substring(start, i));
+                sb.write(r'\');
+                start = i;
+              }
+            }
+            sb..write(value.substring(start))..write('"');
+          }
+        }
       });
     }
     return sb.toString();
   }
 
-  void _parse(String s, String parameterSeparator, String valueSeparator,
+  void _parse(String s, String parameterSeparator, String? valueSeparator,
       bool preserveBackslash) {
     int index = 0;
 
@@ -670,10 +746,11 @@ class _HeaderValue implements HeaderValue {
     String parseValue() {
       int start = index;
       while (!done()) {
-        if (s[index] == " " ||
-            s[index] == "\t" ||
-            s[index] == valueSeparator ||
-            s[index] == parameterSeparator) break;
+        var char = s[index];
+        if (char == " " ||
+            char == "\t" ||
+            char == valueSeparator ||
+            char == parameterSeparator) break;
         index++;
       }
       return s.substring(start, index);
@@ -686,22 +763,26 @@ class _HeaderValue implements HeaderValue {
       index++;
     }
 
-    void maybeExpect(String expected) {
-      if (s[index] == expected) index++;
+    bool maybeExpect(String expected) {
+      if (done() || !s.startsWith(expected, index)) {
+        return false;
+      }
+      index++;
+      return true;
     }
 
     void parseParameters() {
-      var parameters = new HashMap<String, String>();
-      _parameters = new UnmodifiableMapView(parameters);
+      var parameters = _ensureParameters();
 
       String parseParameterName() {
         int start = index;
         while (!done()) {
-          if (s[index] == " " ||
-              s[index] == "\t" ||
-              s[index] == "=" ||
-              s[index] == parameterSeparator ||
-              s[index] == valueSeparator) break;
+          var char = s[index];
+          if (char == " " ||
+              char == "\t" ||
+              char == "=" ||
+              char == parameterSeparator ||
+              char == valueSeparator) break;
           index++;
         }
         return s.substring(start, index).toLowerCase();
@@ -713,26 +794,27 @@ class _HeaderValue implements HeaderValue {
           StringBuffer sb = new StringBuffer();
           index++;
           while (!done()) {
-            if (s[index] == "\\") {
+            var char = s[index];
+            if (char == "\\") {
               if (index + 1 == s.length) {
                 throw new HttpException("Failed to parse header value");
               }
               if (preserveBackslash && s[index + 1] != "\"") {
-                sb.write(s[index]);
+                sb.write(char);
               }
               index++;
-            } else if (s[index] == "\"") {
+            } else if (char == "\"") {
               index++;
-              break;
+              return sb.toString();
             }
-            sb.write(s[index]);
+            char = s[index];
+            sb.write(char);
             index++;
           }
-          return sb.toString();
+          throw new HttpException("Failed to parse header value");
         } else {
           // Parse non-quoted value.
-          var val = parseValue();
-          return val == "" ? null : val;
+          return parseValue();
         }
       }
 
@@ -741,23 +823,18 @@ class _HeaderValue implements HeaderValue {
         if (done()) return;
         String name = parseParameterName();
         skipWS();
-        if (done()) {
+        if (maybeExpect("=")) {
+          skipWS();
+          String value = parseParameterValue();
+          if (name == 'charset' && this is _ContentType) {
+            // Charset parameter of ContentTypes are always lower-case.
+            value = value.toLowerCase();
+          }
+          parameters[name] = value;
+          skipWS();
+        } else if (name.isNotEmpty) {
           parameters[name] = null;
-          return;
         }
-        maybeExpect("=");
-        skipWS();
-        if (done()) {
-          parameters[name] = null;
-          return;
-        }
-        String value = parseParameterValue();
-        if (name == 'charset' && this is _ContentType && value != null) {
-          // Charset parameter of ContentTypes are always lower-case.
-          value = value.toLowerCase();
-        }
-        parameters[name] = value;
-        skipWS();
         if (done()) return;
         // TODO: Implement support for multi-valued parameters.
         if (s[index] == valueSeparator) return;
@@ -769,6 +846,7 @@ class _HeaderValue implements HeaderValue {
     _value = parseValue();
     skipWS();
     if (done()) return;
+    if (s[index] == valueSeparator) return;
     maybeExpect(parameterSeparator);
     parseParameters();
   }
@@ -778,27 +856,30 @@ class _ContentType extends _HeaderValue implements ContentType {
   String _primaryType = "";
   String _subType = "";
 
-  _ContentType(String primaryType, String subType, String charset,
-      Map<String, String> parameters)
+  _ContentType(String primaryType, String subType, String? charset,
+      Map<String, String?> parameters)
       : _primaryType = primaryType,
         _subType = subType,
         super("") {
-    if (_primaryType == null) _primaryType = "";
-    if (_subType == null) _subType = "";
+    // TODO(40614): Remove once non-nullability is sound.
+    String emptyIfNull(String? string) => string ?? "";
+    _primaryType = emptyIfNull(_primaryType);
+    _subType = emptyIfNull(_subType);
     _value = "$_primaryType/$_subType";
-    if (parameters != null) {
-      _ensureParameters();
-      parameters.forEach((String key, String value) {
+    // TODO(40614): Remove once non-nullability is sound.
+    Map<String, String?>? nullableParameters = parameters;
+    if (nullableParameters != null) {
+      var parameterMap = _ensureParameters();
+      nullableParameters.forEach((String key, String? value) {
         String lowerCaseKey = key.toLowerCase();
         if (lowerCaseKey == "charset") {
-          value = value.toLowerCase();
+          value = value?.toLowerCase();
         }
-        this._parameters[lowerCaseKey] = value;
+        parameterMap[lowerCaseKey] = value;
       });
     }
     if (charset != null) {
-      _ensureParameters();
-      this._parameters["charset"] = charset.toLowerCase();
+      _ensureParameters()["charset"] = charset.toLowerCase();
     }
   }
 
@@ -810,7 +891,6 @@ class _ContentType extends _HeaderValue implements ContentType {
     int index = result._value.indexOf("/");
     if (index == -1 || index == (result._value.length - 1)) {
       result._primaryType = result._value.trim().toLowerCase();
-      result._subType = "";
     } else {
       result._primaryType =
           result._value.substring(0, index).trim().toLowerCase();
@@ -825,16 +905,16 @@ class _ContentType extends _HeaderValue implements ContentType {
 
   String get subType => _subType;
 
-  String get charset => parameters["charset"];
+  String? get charset => parameters["charset"];
 }
 
 class _Cookie implements Cookie {
   String _name;
   String _value;
-  DateTime expires;
-  int maxAge;
-  String domain;
-  String path;
+  DateTime? expires;
+  int? maxAge;
+  String? domain;
+  String? path;
   bool httpOnly = false;
   bool secure = false;
 
@@ -856,7 +936,9 @@ class _Cookie implements Cookie {
     _value = newValue;
   }
 
-  _Cookie.fromSetCookieValue(String value) {
+  _Cookie.fromSetCookieValue(String value)
+      : _name = "",
+        _value = "" {
     // Parse the 'set-cookie' header value.
     _parseSetCookieValue(value);
   }
@@ -885,14 +967,6 @@ class _Cookie implements Cookie {
       return s.substring(start, index).trim();
     }
 
-    void expect(String expected) {
-      if (done()) throw new HttpException("Failed to parse header value [$s]");
-      if (s[index] != expected) {
-        throw new HttpException("Failed to parse header value [$s]");
-      }
-      index++;
-    }
-
     void parseAttributes() {
       String parseAttributeName() {
         int start = index;
@@ -950,6 +1024,7 @@ class _Cookie implements Cookie {
   String toString() {
     StringBuffer sb = new StringBuffer();
     sb..write(_name)..write("=")..write(_value);
+    var expires = this.expires;
     if (expires != null) {
       sb..write("; Expires=")..write(HttpDate.format(expires));
     }
@@ -989,7 +1064,7 @@ class _Cookie implements Cookie {
     ];
     if (newName == null) throw new ArgumentError.notNull("name");
     for (int i = 0; i < newName.length; i++) {
-      int codeUnit = newName.codeUnits[i];
+      int codeUnit = newName.codeUnitAt(i);
       if (codeUnit <= 32 ||
           codeUnit >= 127 ||
           separators.indexOf(newName[i]) >= 0) {
